// Copyright 2006-2007 Pierre Evenou
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Library General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
// pierre.evenou@univ-nantes.fr

#ifndef ANGLE_HPP_
#define ANGLE_HPP_

#include <cmath>
#include <iostream>
#include <vector>

namespace Mojette
{

/**
 * \brief Mojette angle.
 * 
 * Projection direction defined by p, q with :
 * - gcd(p ,q) = 1
 * - q \f$\geq\f$ 0
 * - if q = 0 \f$\Rightarrow\f$ p = 1
 * - if p = 0 \f$\Rightarrow\f$ q = 1 
 */
class Angle
{
public:

    /**
     * Construct angle with p = 0 and q = 1 as default values.
     * 
     * According to the mojette angle definition, 
     * all given p and q values are reduced 
     * (eg : (2, 2) = (1, 1)).
     * 
     * \param p : the p value.
     * \param q : the q value.
     */
    Angle(short p = 0, unsigned short q = 1);

    /**
     * Equal operator.
     * 
     * \return true if the two angles are equal otherwise false.
     */
    bool operator ==(const Angle& a);
    
    /**
     * Not equals operator.
     * 
     * \return true if the two angles are not equal otherwise false.
     */
    bool operator !=(const Angle& a);
    
    /**
     * Greater than operator (theta is compared).
     * 
     * \return true this angle is greater than the given
     * angle otherwise false.
     */
    bool operator >(const Angle& a);
    
    /**
     * Lesser than operator (theta is compared).
     * 
     * \return true this angle is lesser than the given 
     * angle otherwise false.
     */
    bool operator <(const Angle& a);

    /**
     * Getter for the p value.
     * 
     * \return the p value.
     */
    short getP() const;
    
    /**
     * Setter for the p value.
     * 
     * \param p : the p value.
     */
    void setP(const short p);
    
    /**
     * Getter for the q value.
     * 
     * \return the q value.
     */
    unsigned short getQ() const;
    
    /**
     * Setter for the q value.
     * 
     * \param q : the q value.
     */
    void setQ(const unsigned short q);
    
    /**
     * Getter for the tetha value.
     * 
     * \return tetha = atan(\f$\frac{q}{p}\f$).
     */
    double getTheta() const;
    
private:
    
    short mP;
    
    unsigned short mQ;

    void reduce();
};

inline short
Angle::getP() const
{
    return mP;
}

inline void
Angle::setP(const short p)
{
    mP = p;
    reduce();
}

inline unsigned short
Angle::getQ() const
{
    return mQ;
}

inline void
Angle::setQ(const unsigned short q)
{
    mQ = q;
    reduce();
}

inline double
Angle::getTheta() const
{
    return (std::atan2((double)mQ, (double)mP));
}

struct SortByAngle
{
    bool operator ()(const Angle& a1, const Angle& a2) const;
};

}

#endif
