// Copyright 2006-2007 Pierre Evenou
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Library General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
// pierre.evenou@univ-nantes.fr

#ifndef DIRAC_TRANSFORM_HPP_
#define DIRAC_TRANSFORM_HPP_

#include <vector>

#include "Angle.hpp"
#include "Operator.hpp"
#include "Transform.hpp"

namespace Mojette
{

/**
 * \brief Dirac Mojette Transform.
 * 
 * Class for Dirac Mojette transform of rectangular support 
 * defined by L and K respectively height and width of the support.
 */
template <class T, class U, int type>
class DiracTransform : public Transform<T, U, type>
{
    
public :

    /**
     * \copydoc Transform::Transform(unsigned int L, unsigned int K)
     */
    DiracTransform(unsigned int L = 1, unsigned int K = 1);
    
    /**
     * Do a forward Dirac transform of the given pixels.
     * 
     * \param angles : the set of angles.
     * \param pixels : the pixels to transform.
     * 
     * \return the projections.
     */
    virtual std::vector<std::vector<U> >* forward(const std::vector<Angle>& angles, const std::vector<T>& pixels);
   
    /**
     * Do an inverse Dirac transform of the given projections.
     * 
     * \param angles : the set of angles.
     * \param projections : the projections to transform.
     * 
     * \return the reconstruted pixels.
     */   
    virtual std::vector<T>* inverse(const std::vector<Angle>& angles, std::vector<std::vector<U> >& projections);
    
    /**
     * Give the size of a Dirac projection for a given angle.
     * 
     * \param angle : the angle.
     * 
     * \return the size of the corresponding projection.
     */
    unsigned int projectionSize(const Angle& a);
    
    /**
     * Give the index of a bin on a Dirac projection for a given angle 
     * and co-ordinates on a support.
     * 
     * \param angle : the angle.
     * \param l : the line.
     * \param l : the column.
     * 
     * \return the index of the corresponding bin.
     */
    unsigned int binIndex(const Angle& a, unsigned int l, unsigned int k);

};

template <class T, class U, int type>
DiracTransform<T, U, type>::DiracTransform(unsigned int L, unsigned int K) 
: Transform<T, U, type>(L, K){}

template <class T, class U, int type>
std::vector<std::vector<U> >*
DiracTransform<T, U, type>::forward(const std::vector<Angle>& angles, const std::vector<T>& pixels)
{
    std::vector<std::vector<U> >* projections = 
        new std::vector<std::vector<U> >();
    
    // TODO : throw exception.
    if (pixels.size() != Transform<T, U, type>::mL * Transform<T, U, type>::mK)
        return new std::vector<std::vector<U> >();
       
    std::vector<unsigned int> offsets;
    for (unsigned int i = 0; i < angles.size(); i ++)
    {
        const Angle& a = angles[i];
        offsets.push_back(a.getP() < 0 ? (Transform<T, U, type>::mL - 1) * a.getP() : 0);
        projections->push_back(std::vector<U>(abs(a.getP()) * (Transform<T, U, type>::mL - 1) + 
            a.getQ() * (Transform<T, U, type>::mK - 1) + 1));
    }
    
    Operator<U, type> op;
    for (unsigned int l = 0; l < Transform<T, U, type>::mL; l ++)
    {
        for (unsigned int k = 0; k < Transform<T, U, type>::mK; k ++)
        {
            U px = (U)pixels[l * Transform<T, U, type>::mK + k];
            
            for (unsigned int i = 0; i < projections->size(); i ++)
            {
                Angle a = angles[i];
                op.forward(projections->operator[](i)
                    [k * a.getQ() + l * a.getP() - offsets[i]], px); 
            }
        }
    }
    
    return projections;
}

struct BinIndex
{
    unsigned int p;
    unsigned long b;
    
    BinIndex(unsigned int p, unsigned long b) : p(p), b(b) {}
};

template <class T, class U, int type>
std::vector<T>*
DiracTransform<T, U, type>::inverse(const std::vector<Angle>& angles, std::vector<std::vector<U> >& projections)
{   
    // TODO : throw exception.
    if (projections.size() != angles.size())
        return new std::vector<T>();
        
    std::vector<unsigned int> offsets;
    std::vector<std::vector<unsigned long> > pu;
    std::vector<std::vector<unsigned long long> > pd;
    
    for (unsigned int i = 0; i < angles.size(); i ++)
    {
        const Angle& a = angles[i];
        offsets.push_back(a.getP() < 0 ? (Transform<T, U, type>::mL - 1) * a.getP() : 0);
        
        int p_size = abs(a.getP()) * (Transform<T, U, type>::mL - 1) + a.getQ() * (Transform<T, U, type>::mK - 1) + 1;
        pu.push_back(std::vector<unsigned long>(p_size));
        pd.push_back(std::vector<unsigned long long>(p_size));
    }
    
    unsigned long long d = 0;
    for (unsigned int l = 0; l < Transform<T, U, type>::mL; l ++)
    {
        for (unsigned int k = 0; k < Transform<T, U, type>::mK; k ++)
        {
            for (unsigned int i = 0; i < angles.size(); i ++)
            {
                const Angle& a = angles[i];
                unsigned int b_idx = k * a.getQ() + l * a.getP() - offsets[i];
                pu[i][b_idx] += 1;
                pd[i][b_idx] += d;
            }
            d ++;
        }
    }
    
    std::vector<BinIndex> univocs;
    for (unsigned int i = 0; i < angles.size(); i ++)
        for (unsigned int j = 0; j < pu[i].size(); j ++)
            if (pu[i][j] == 1)
                univocs.push_back(BinIndex(i, j));
    
    std::vector<T>* pixels = new std::vector<T>(Transform<T, U, type>::mL * Transform<T, U, type>::mK);
    unsigned int nb_pixels = 0;
    unsigned int curr_univoc = 0;
    Operator<U, type> op;
    while (nb_pixels ++ < Transform<T, U, type>::mL * Transform<T, U, type>::mK)
    {
        BinIndex& u = univocs[curr_univoc ++];
        while (pu[u.p][u.b] != 1)
            u = univocs[curr_univoc ++];
        
        U bin = projections[u.p][u.b];
        
        unsigned long long diet = pd[u.p][u.b];
        unsigned int l = (unsigned int)(diet / Transform<T, U, type>::mK);
        unsigned int k = (unsigned int)(diet - l * Transform<T, U, type>::mK);
        
        pixels->operator[](l * Transform<T, U, type>::mK + k) = (T)bin;
        
        for (unsigned int i = 0; i < angles.size(); i ++)
        {
            const Angle& a = angles[i];
            unsigned int b_idx = k * a.getQ() + l * a.getP() - offsets[i];
            
            op.inverse(projections[i][b_idx], bin);
            
            pd[i][b_idx] -= diet;
            
            if (--pu[i][b_idx] == 1)
                univocs.push_back(BinIndex(i, b_idx));
        }
    }
    return pixels;
}

template <class T, class U, int type>
inline unsigned int 
DiracTransform<T, U, type>::projectionSize(const Angle& a)
{
    return (abs(a.getP()) * (Transform<T, U, type>::mL - 1) + a.getQ() * (Transform<T, U, type>::mK - 1) + 1);
}

template <class T, class U, int type>
inline unsigned int 
DiracTransform<T, U, type>::binIndex(const Angle& a, unsigned int l, unsigned int k)
{
    unsigned int offset = a.getP() < 0 ? (Transform<T, U, type>::mL - 1) * a.getP() : 0;
    return (k * a.getQ() + l * a.getP() - offset);
}

}

#endif
