// Copyright 2006-2007 Pierre Evenou
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Library General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
// pierre.evenou@univ-nantes.fr

#ifndef OPERATOR_HPP_
#define OPERATOR_HPP_

namespace Mojette
{

/**
 * \brief Operator type.
 * 
 * Projection operator type. 
 */
enum OperatorType
{
    XOR,
    ADD,
    MOD
};

// Use meta programing to replace switch statement.
/**
 * \brief Mojette projection operator.
 */
template <class T, int type>
class Operator
{
public :

    /**
     * Project pixel on bin according to operator type.
     * 
     * \param bin : the bin value.
     * \param pixel : the pixel value.
     */
    void forward(T& bin, const T& pixel);
    
    /**
     * Retro project pixel on bin according to operator type.
     * 
     * \param bin : the bin value.
     * \param pixel : the pixel value.
     */
    void inverse(T& bin, const T& pixel);
};

// Default statement : do nothing
template <class T, int type>
inline void
Operator<T, type>::forward(T& bin, const T& pixel)
{}

template <class T, int type>
inline void
Operator<T, type>::inverse(T& bin, const T& pixel)
{}

/**
 * \brief Mojette XOR projection operator.
 */
template <class T>
class Operator<T, XOR>
{
public :

    /**
     * Project pixel on bin according to operator type.
     * bin ^= pixel
     * 
     * \param bin : the bin value.
     * \param pixel : the pixel value.
     */
    void forward(T& bin, const T& pixel);
    
    /**
     * Retro project pixel on bin according to operator type.
     * * bin ^= pixel
     * 
     * \param bin : the bin value.
     * \param pixel : the pixel value.
     */
    void inverse(T& bin, const T& pixel);
};

template <class T>
inline void
Operator<T, XOR>::forward(T& bin, const T& pixel)
{
    bin ^= pixel;
}

template <class T>
inline void
Operator<T, XOR>::inverse(T& bin, const T& pixel)
{
    bin ^= pixel;
}

/**
 * \brief Mojette ADD projection operator.
 */
template <class T>
class Operator<T, ADD>
{
public :

    /**
     * Project pixel on bin according to operator type.
     * bin += pixel
     * 
     * \param bin : the bin value.
     * \param pixel : the pixel value.
     */
    void forward(T& bin, const T& pixel);
    
    /**
     * Retro project pixel on bin according to operator type.
     * * bin -= pixel
     * 
     * \param bin : the bin value.
     * \param pixel : the pixel value.
     */
    void inverse(T& bin, const T& pixel);
};

template <class T>
inline void
Operator<T, ADD>::forward(T& bin, const T& pixel)
{
    bin += pixel;
}

template <class T>
inline void
Operator<T, ADD>::inverse(T& bin, const T& pixel)
{
    bin -= pixel;
}

/*
// XXX : BUGGY
template <class T>
class Operator<T, MOD>
{
public :

    Operator(T modulus);
    
    T forward(const T& t1, const T& t2);
    
    T inverse(const T& t1, const T& t2);
    
private :

    int mModulus;
};

template <class T, MOD>
Operator(T modulus) : mModulus(modulus)
{}

template <class T, MOD>
inline T
Operator<T, MOD>::forward(const T& t1, const T& t2)
{
    return (t1 + t2) % modulus;
}

template <class T, MOD>
inline T
Operator<T, MOD>::inverse(const T& t1, const T& t2)
{
    return (t1 - t2) % modulus;
}
*/
}

#endif
