// Copyright 2006-2007 Pierre Evenou
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Library General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
// pierre.evenou@univ-nantes.fr

#ifndef TRANSFORM_HPP_
#define TRANSFORM_HPP_

#include <vector>
#include <cmath>

namespace Mojette
{

class Angle;

/**
 * \brief Mojette Transform.
 * 
 * Base class for Mojette transform of rectangular support 
 * defined by L and K respectively height and width of the support.
 */
template <class T, class U, int type>
class Transform
{

public :

    /**
     * Construct transform.
     * 
     * L and K values gives the height and width of the support used.
     * 
     * \param L : the height of the support.
     * \param K : the width of the support.
     */
    Transform(unsigned int L = 1, unsigned int K = 1);

    virtual ~Transform();
    
    /**
     * Getter for the L value.
     * 
     * \return the L value.
     */
    const unsigned int& getL() const;
    
    /**
     * Getter for the K value.
     * 
     * \return the K value.
     */
    const unsigned int& getK() const;
    
    /**
     * Setter for the L value.
     * 
     * \param l : the L value.
     */
    void setL(unsigned int l);
    
    /**
     * Setter for the K value.
     * 
     * \param k : the K value.
     */
    void setK(unsigned int k);
    
    /**
     * Give the reconstructibility of a set of angles for the transform.
     * Use Katz lemma.
     * 
     * \param angles : the set of angles.
     * 
     * \return true if the given angle set is reconstructible false otherwise.
     */
    bool isReconstructible(const std::vector<Angle>& angles);
    
    virtual std::vector<std::vector<U> >* forward(
	    const std::vector<Angle>& angles, 
	    const std::vector<T>& pixels) = 0;
   
    virtual std::vector<T>* inverse(
	    const std::vector<Angle>& angles, 
	    std::vector<std::vector<U> >& projections) = 0;
    
    virtual unsigned int projectionSize(const Angle& a) = 0;
    
    virtual unsigned int binIndex(const Angle& a, unsigned int l, 
	    unsigned int k) = 0;
    
protected :
    
    unsigned int mL;
    
    unsigned int mK;
};

template <class T, class U, int type>
Transform<T, U, type>::Transform(unsigned int L, unsigned int K) : mL(L), mK(K) 
{}

template <class T, class U, int type>
Transform<T, U, type>::~Transform() 
{}

template <class T, class U, int type>
inline const unsigned int&
Transform<T, U, type>::getL() const
{
    return mL;
}

template <class T, class U, int type>
inline const unsigned int&
Transform<T, U, type>::getK() const
{
    return mK;
}

template <class T, class U, int type>
inline void
Transform<T, U, type>::setL(unsigned int l)
{
    mL = l;
}

template <class T, class U, int type>
inline void
Transform<T, U, type>::setK(unsigned int k)
{
    mK = k;
}

template <class T, class U, int type>
inline bool
Transform<T, U, type>::isReconstructible(const std::vector<Angle>& angles)
{
    unsigned int sum_p = 0;
    unsigned int sum_q = 0;
    
    for (int i = 0; i < angles.size(); i ++)
    {
        sum_p += angles[i].getP();
        sum_q += angles[i].getQ();
    }
    
    return sum_p >= mK || sum_q >= mL;
}

}

#endif
